#include <engine/3d/shapes.h>
#include <engine/3d/camera.h>
#include <engine/display.h>

/*
 * add_quad:
 *	Performs transformation of 3D coords into
 *	screen coords for quads.
 */
add_quad(p_ot, p_quad, p_global_pos, p_rottrans, use_pivot_local)
unsigned long *p_ot;
struct Quad *p_quad;
SVECTOR *p_global_pos; /* player's global position */
MATRIX *p_rottrans;
int use_pivot_local;
{
	int i;

	MATRIX		model_rot;
	SVECTOR		d_Pw = {0, 0, 0};			/* Pw' (world position transform) */
	SVECTOR		dmy[4];
	long		dmy_long;
	long		p, otz, flg;

	/* apply world_position transform first! */
	if(!use_pivot_local)
	{
		setVector(&d_Pw, p_quad->global_pos.vx + p_global_pos->vx,
				p_quad->global_pos.vy + p_global_pos->vy,
				p_quad->global_pos.vz + p_global_pos->vz);
	}

	PushMatrix();

	if(p_quad->is_coplanar)
	{
		dmy[0] = dmy[1] = dmy[2] = dmy[3] = d_Pw;

		dmy[0].vx -= p_quad->size; dmy[0].vz -= p_quad->size; /* top left */
		dmy[1].vx += p_quad->size; dmy[1].vz -= p_quad->size; /* top right */
		dmy[2].vx -= p_quad->size; dmy[2].vz += p_quad->size; /* bottom left */
		dmy[3].vx += p_quad->size; dmy[3].vz += p_quad->size; /* bottom right */
	}
	else
	{
		/* read model coords, then translate */
		for(i = 0; i < 4; i++)
		{
			dmy[i] = p_quad->global_coords.v[i];

			dmy[i].vx += d_Pw.vx;
			dmy[i].vy += d_Pw.vy;
			dmy[i].vz += d_Pw.vz;
		}
	}

	/* first, rotate the quad in model space (XYZ), then translate */
	if(use_pivot_local)
	{
		VECTOR out[4]; /* FIXME: should just be one VECTOR, not 4 */

		/* FIXME: RotMatrix_gte() can cause crashes for some reason.. dunno why */
		RotMatrix(&p_quad->rotation_model, &model_rot);

		for(i = 0; i < 4; i++)
		{
			ApplyMatrix(&model_rot, &dmy[i], &out[i]);
			setVector(&dmy[i], out[i].vx, out[i].vy, out[i].vz);

			/*
			 * FIXME: can move this out of use_pivot_local block and don't apply on
			 * world position transform.
			 */
			dmy[i].vx += p_global_pos->vx;
			dmy[i].vy += p_global_pos->vy;
			dmy[i].vz += p_global_pos->vz;
		}
	}

	RotMatrix(&p_quad->rotation, p_rottrans);
	TransMatrix(p_rottrans, &p_quad->translation);

	SetRotMatrix(p_rottrans);
	SetTransMatrix(p_rottrans);

	otz = RotAverage4(&dmy[0], &dmy[1], &dmy[2], &dmy[3],
		(long *)(&p_quad->prim.x0), (long *)(&p_quad->prim.x1),
		(long *)(&p_quad->prim.x2), (long *)(&p_quad->prim.x3),
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_quad->prim);
	}

	PopMatrix();
}

/*
 * add_cuboid:
 *	Performs transformation of 3D coords into
 *	screen coords for cuboids. Size in each
 *	dimension is determined via size_x, size_y,
 *	and size_z members.
 */
add_cuboid(p_ot, p_cuboid, p_camera, p_rottrans)
unsigned long *p_ot;
struct Cuboid *p_cuboid;
struct Camera3D *p_camera;
MATRIX *p_rottrans;
{

	/*
	 * FIXME: try replacing every instance of RotAverage3() here
	 * with RotTransPers()
	 */

	int		i;
	SVECTOR		dmy0, dmy1, dmy2, dmy3;
	long		dmy_long;
	long		p, otz, flg;

	SVECTOR		face_left, face_right, face_top, face_bottom;

	/* left */
	setVector(&face_left, (p_cuboid->offset.vx - p_cuboid->size_x/2) - p_camera->offset.vx,
			p_cuboid->offset.vy - p_camera->offset.vy,
			p_cuboid->offset.vz - p_camera->offset.vz);

	/* right */
	setVector(&face_right, (p_cuboid->offset.vx + p_cuboid->size_x/2) - p_camera->offset.vx,
			p_cuboid->offset.vy - p_camera->offset.vy,
			p_cuboid->offset.vz - p_camera->offset.vz);

	/* top */
	setVector(&face_top, p_cuboid->offset.vx - p_camera->offset.vx,
			(p_cuboid->offset.vy + p_cuboid->size_y/2) - p_camera->offset.vy,
			p_cuboid->offset.vz - p_camera->offset.vz);

	/* bottom */
	setVector(&face_bottom, p_cuboid->offset.vx - p_camera->offset.vx,
			(p_cuboid->offset.vy - p_cuboid->size_y/2) - p_camera->offset.vy,
			p_cuboid->offset.vz - p_camera->offset.vz);

	PushMatrix();

	RotMatrix(&p_cuboid->rotation, p_rottrans);
	TransMatrix(p_rottrans, &p_cuboid->translation);

	SetRotMatrix(p_rottrans);
	SetTransMatrix(p_rottrans);

	/***** face_left *****/
	dmy0 = dmy1 = dmy2 = dmy3 = face_left;

	/* do not do criss-cross pattern, for they are LINE_F4s */
	dmy0.vy += p_cuboid->size_y/2; dmy0.vz += p_cuboid->size_z/2;
	dmy1.vy += p_cuboid->size_y/2; dmy1.vz -= p_cuboid->size_z/2;
	dmy2.vy -= p_cuboid->size_y/2; dmy2.vz -= p_cuboid->size_z/2;
	dmy3.vy -= p_cuboid->size_y/2; dmy3.vz += p_cuboid->size_z/2;

	otz = RotAverage4(&dmy0, &dmy1, &dmy2, &dmy3,
		(long *)(&p_cuboid->left.x0), (long *)(&p_cuboid->left.x1),
		(long *)(&p_cuboid->left.x2), (long *)(&p_cuboid->left.x3),
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_cuboid->left);
	}

	/* do missing line to complete the face */
	otz = RotAverage3(&dmy0, &dmy3, &dmy0,
		(long *)(&p_cuboid->left_.x0), (long *)(&p_cuboid->left_.x1), &dmy_long,
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_cuboid->left_);
	}

	/***** face_right *****/
	dmy0.vx += p_cuboid->size_x; /* shift to the right side */
	dmy1.vx += p_cuboid->size_x;
	dmy2.vx += p_cuboid->size_x;
	dmy3.vx += p_cuboid->size_x;

	otz = RotAverage4(&dmy0, &dmy1, &dmy2, &dmy3,
		(long *)(&p_cuboid->right.x0), (long *)(&p_cuboid->right.x1),
		(long *)(&p_cuboid->right.x2), (long *)(&p_cuboid->right.x3),
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_cuboid->right);
	}

	otz = RotAverage3(&dmy0, &dmy3, &dmy0,
		(long *)(&p_cuboid->right_.x0), (long *)(&p_cuboid->right_.x1), &dmy_long,
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_cuboid->right_);
	}

	/***** face_top *****/
	dmy0 = dmy1 = dmy2 = dmy3 = face_top;

	dmy0.vx -= p_cuboid->size_x/2; dmy0.vz -= p_cuboid->size_z/2;
	dmy1.vx += p_cuboid->size_x/2; dmy1.vz -= p_cuboid->size_z/2;
	dmy2.vx += p_cuboid->size_x/2; dmy2.vz += p_cuboid->size_z/2;
	dmy3.vx -= p_cuboid->size_x/2; dmy3.vz += p_cuboid->size_z/2;

	otz = RotAverage4(&dmy0, &dmy1, &dmy2, &dmy3,
		(long *)(&p_cuboid->top.x0), (long *)(&p_cuboid->top.x1),
		(long *)(&p_cuboid->top.x2), (long *)(&p_cuboid->top.x3),
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_cuboid->top);
	}

	otz = RotAverage3(&dmy0, &dmy3, &dmy0,
		(long *)(&p_cuboid->top_.x0), (long *)(&p_cuboid->top_.x1), &dmy_long,
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_cuboid->top_);
	}

	/***** face_bottom *****/
	dmy0.vy -= p_cuboid->size_y; /* shift to the bottom side */
	dmy1.vy -= p_cuboid->size_y;
	dmy2.vy -= p_cuboid->size_y;
	dmy3.vy -= p_cuboid->size_y;

	otz = RotAverage4(&dmy0, &dmy1, &dmy2, &dmy3,
		(long *)(&p_cuboid->bottom.x0), (long *)(&p_cuboid->bottom.x1),
		(long *)(&p_cuboid->bottom.x2), (long *)(&p_cuboid->bottom.x3),
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_cuboid->bottom);
	}

	otz = RotAverage3(&dmy0, &dmy3, &dmy0,
		(long *)(&p_cuboid->bottom_.x0), (long *)(&p_cuboid->bottom_.x1), &dmy_long,
		&p, &flg);

	if(otz > MIN_OTZ && otz < game_otsize)
	{
		AddPrim(p_ot+otz, &p_cuboid->bottom_);
	}

	PopMatrix();
}

/*
 * set_circle3d_color:
 *	Sets the color of a 3D circle.
 */
void
set_circle3d_color(p_circle, r, g, b)
struct Circle3D *p_circle;
char r, g, b;
{
	int i;

	for(i = 0; i < CIRCLE_LINEF2; i++)
	{
		setRGB0(&p_circle->lines[i], r, g, b);
	}
}

/*
 * update_circle3d:
 *	Updates the model coordinates of a 3D circle.
 *	Make sure you call this during initialization so
 *	that the circle has correct model coordinates!
 */
void
update_circle3d(p_circle)
struct Circle3D *p_circle;
{
	int		i, j;
	const int	unit = 4096/CIRCLE_LINEF2;

	for(i = 0; i < CIRCLE_LINEF2; i+=3)
	{
		int	angle = i*unit;

		switch(p_circle->orientation)
		{
			default:
			case CIRCLE_HORIZONTAL:
				for(j = 0; j < 3; j++)
				{
					p_circle->vertex[i+j].vx = (rcos(angle+unit*j)*p_circle->ent.scale)>>12;

					p_circle->vertex[i+j].vy = 0;
					p_circle->vertex[i+j].vz = (rsin(angle+unit*j)*p_circle->ent.scale)>>12;
				}
				break;
			case CIRCLE_VERTICAL:
				for(j = 0; j < 3; j++)
				{
					p_circle->vertex[i+j].vx = (rcos(angle+unit*j)*p_circle->ent.scale)>>12;

					p_circle->vertex[i+j].vy = (rsin(angle+unit*j)*p_circle->ent.scale)>>12;
					p_circle->vertex[i+j].vz = 0;
				}
				break;
			case CIRCLE_CUSTOM:
				for(j = 0; j < 3; j++)
				{
					int _cos = rcos(angle+unit*j);
					int _sin = rsin(angle+unit*j);

					VECTOR b_hat = {(_cos*p_circle->unit[0].vx)>>6, (_cos*p_circle->unit[0].vy)>>6, (_cos*p_circle->unit[0].vz)>>6};
					VECTOR c_hat = {(_sin*p_circle->unit[1].vx)>>6, (_sin*p_circle->unit[1].vy)>>6, (_sin*p_circle->unit[1].vz)>>6};

					p_circle->vertex[i+j].vx = ((b_hat.vx+c_hat.vx)*p_circle->ent.scale)>>12;
					p_circle->vertex[i+j].vy = ((b_hat.vy+c_hat.vy)*p_circle->ent.scale)>>12;
					p_circle->vertex[i+j].vz = ((b_hat.vz+c_hat.vz)*p_circle->ent.scale)>>12;
				}
				break;
		}
	}
}

/*
 * set_cylinder:
 *	Initializes a cylinder.
 */
void
set_cylinder(p_cylinder, r, x0, y0, z0, x1, y1, z1)
struct Cylinder *p_cylinder;
int r;
short x0, y0, z0, x1, y1, z1;
{
	setVector(&p_cylinder->p[0], x0, y0, z0);
	setVector(&p_cylinder->p[1], x1, y1, z1);

	p_cylinder->h = psqrt((ppow(x0-x1, 2)>>6) + (ppow(y0-y1, 2)>>6) + (ppow(z0-z1, 2)>>6));
	p_cylinder->r = r;
}

/*
 * cylinder_update:
 *	Updates members that are dependent on the endpoints of the cylinder.
 */
void
cylinder_update(p_cylinder)
struct Cylinder *p_cylinder;
{
	p_cylinder->h = psqrt((ppow(p_cylinder->p[0].vx-p_cylinder->p[1].vx, 2)>>6) +
		(ppow(p_cylinder->p[0].vy-p_cylinder->p[1].vy, 2)>>6) +
		(ppow(p_cylinder->p[0].vz-p_cylinder->p[1].vz, 2)>>6));
}

/*
 * cylinder_gfx_update:
 *	Updates members of a CylinderGFX struct.
 */
void
cylinder_gfx_update(p_gfx)
struct CylinderGFX *p_gfx;
{
	cylinder_update(&p_gfx->cyl);
	p_gfx->cyl.h = !p_gfx->cyl.h ? 1 : p_gfx->cyl.h;
	p_gfx->unit.vx = ((p_gfx->cyl.p[1].vx<<6)-(p_gfx->cyl.p[0].vx<<6))/p_gfx->cyl.h;
	p_gfx->unit.vy = ((p_gfx->cyl.p[1].vy<<6)-(p_gfx->cyl.p[0].vy<<6))/p_gfx->cyl.h;
	p_gfx->unit.vz = ((p_gfx->cyl.p[1].vz<<6)-(p_gfx->cyl.p[0].vz<<6))/p_gfx->cyl.h;

	p_gfx->endc[0].ent.scale = p_gfx->cyl.r;
	p_gfx->endc[1].ent.scale = p_gfx->cyl.r;
}

#define DEFAULT_CIRCLE_SIZE 64

/*
 * set_cylinder_gfx:
 *	Initializes the graphics for a cylinder.
 */
void
set_cylinder_gfx(p_gfx, r)
struct CylinderGFX *p_gfx;
{
	int i, j;

	ent_zero(&p_gfx->ent, 0, 1);
	p_gfx->cyl.r = r;

	for(i = 0; i < 2; i++)
	{
		setTile(p_gfx->endp+i);
		setWH(p_gfx->endp+i, 2, 2);
		setRGB0(p_gfx->endp+i, 0, 0, 0);

		ent_zero(&p_gfx->endc[i].ent, 0, r);
		p_gfx->endc[i].orientation = CIRCLE_VERTICAL;
		update_circle3d(p_gfx->endc+i); /* FIXME: horizontal orientation is temporary, we'll use custom eventually */

		for(j = 0; j < CIRCLE_LINEF2; j++)
		{
			setLineF2(&p_gfx->endc[i].lines[j]);
			setRGB0(&p_gfx->endc[i].lines[j], 0, 0, 0);
			p_gfx->endc[i].visual = SOLID;
			p_gfx->endc[i].orientation = CIRCLE_VERTICAL;
		}
	}

	/* now initialize the body of the cylinder's body lines */
	for(i = 0; i < CYLINDERGFX_BODY_NLINES; i++)
	{
		for(j = 0; j < CYLINDERGFX_BODY_NSEGMENTS; j++)
		{
			setLineF2(&p_gfx->body[i][j]);
			setRGB0(&p_gfx->body[i][j], 0, 0, 0);
		}
	}
}

/*
 * set_cylinder_gfx_to_camera:
 *	Registers the entities of a cylinder to a camera.
 */
void
set_cylinder_gfx_to_camera(p_gfx, p_camera)
struct CylinderGFX *p_gfx;
struct Camera3D *p_camera;
{
	int i;

	camera3d_reg_ent(p_camera, &p_gfx->ent);

	for(i = 0; i < 2; i++)
	{
		camera3d_reg_ent(p_camera, &p_gfx->endc[i].ent);
	}
}

set_cylinder_gfx_color(p_gfx, r, g, b)
struct CylinderGFX *p_gfx;
unsigned char r, g, b;
{
	int i, j;
	for(i = 0; i < CYLINDERGFX_BODY_NLINES; i++)
	{
		for(j = 0; j < CYLINDERGFX_BODY_NSEGMENTS; j++)
		{
			setRGB0(&p_gfx->body[i][j], r, g, b);
		}
	}
}

